from __future__ import annotations
from abc import ABC, abstractmethod


class Type(ABC):
    """ Abstract base class for all Cowait types. """

    name: str = None

    @abstractmethod
    def validate(self, value: any, name: str) -> None:
        """ Validates a value as this type. Raises ValueError if invalid """
        raise NotImplementedError()

    def serialize(self, value: Type) -> object:
        """ Returns a JSON-serializable representation of the value """
        return value

    def deserialize(self, value: any) -> Type:
        """ Deserializes a JSON representation of a value """
        return value

    def describe(self) -> any:
        """ Creates a JSON-serializable type description for this type """
        if self.name is None:
            raise TypeError('Unnamed type')
        return self.name



from cowait.tasks import TASK_LOG
from cowait.worker.logger import Logger
from IPython.core.display import display, HTML


class HTMLLogger(Logger):
    def log(self, type: str, msg: dict) -> None:
        if type == TASK_LOG:
            self.log_output(**msg)

    def log_output(self, id, file, data, **msg):
        display(HTML(f'<pre>{id}: {data}</pre>'))



from cowait.tasks.shell import ShellTask


class DaskWorker(ShellTask):
    async def run(self, scheduler: str, **inputs):
        command = f'dask-worker {scheduler}'
        return await super().run(command=command, **inputs)



from cowait.tasks.shell import ShellTask


class DaskScheduler(ShellTask):
    async def run(self, **inputs):
        return await super().run(
            command='dask-scheduler',
            **inputs,
        )



import socket
from .const import WS_PATH, QUERY_TOKEN


def get_local_ip():
    hostname = socket.gethostname()
    return socket.gethostbyname(hostname)


def get_local_url():
    local_ip = get_local_ip()
    return f'ws://{local_ip}/{WS_PATH}'


def get_remote_url(host, token):
    return f'ws://{host}/{WS_PATH}?{QUERY_TOKEN}={token}'



from __future__ import annotations
import asyncio
from concurrent.futures import Future
from cowait.types import type_from_description
from .errors import TaskError, StoppedError
from .instance import TaskInstance
from .definition import TaskDefinition
from .status import WAIT, WORK, FAIL, DONE, STOP


class RemoteTask(TaskInstance):
    def __init__(self, taskdef: TaskDefinition, cluster):
        kwargs = taskdef.serialize()
        super().__init__(**kwargs)
        self.conn = None
        self.nonce = 0
        self.cluster = cluster
        self.future = Future()
        self.awaitable = asyncio.wrap_future(self.future)

    def __await__(self):
        return self.awaitable.__await__()

    @property
    def done(self) -> bool:
        return self.future.done()

    def destroy(self) -> None:
        self.cluster.destroy(self.id)

    def set_status(self, status: str) -> None:
        # sanity checks
        if self.status == FAIL and status == DONE:
            raise RuntimeError('Cant complete a failed task')

        if self.status == DONE and status == FAIL:
            raise RuntimeError('Cant fail a completed task')

        if status == STOP and not self.future.done():
            self.future.set_exception(StoppedError(f'Remote task {self.id} was stopped'))

        # update status
        self.status = status

    def set_error(self, error: str) -> None:
        self.set_status(FAIL)
        self.error = error
        if not self.future.done():
            self.future.set_exception(TaskError(error))

    def set_result(self, result: any, result_type: any = 'any') -> None:
        # unpack type & deserialize result
        result_type = type_from_description(result_type)
        result = result_type.deserialize(result)

        self.set_status(DONE)
        self.result = result
        if not self.future.done():
            self.future.set_result(result)

    async def wait_for_scheduling(self) -> None:
        pass

    async def wait_for_init(self, timeout=30) -> None:
        if self.status != WAIT:
            raise RuntimeError(f'Cant await task with status {self.status}')

        await self.wait_for_scheduling()

        slept = 0
        interval = 0.2
        while True:
            if self.status == WORK:
                return
            if self.status == FAIL:
                raise RuntimeError(
                    f'Awaited task failed with error: {self.error}')

            if slept > timeout:
                raise TimeoutError('Task took to long to initialize')

            await asyncio.sleep(interval)
            slept += interval

    async def call(self, method, args={}) -> any:
        if self.status == WAIT:
            await self.wait_for_init()
        elif self.status != WORK:
            raise RuntimeError(
                f'RPC is only available when status = WORK, was {self.status}. '
                f'Attempted to call {method}')

        return await self.conn.rpc.call(method, args)

    async def stop(self) -> None:
        if self.status == STOP:
            return
        await self.call('stop')

    def logs(self):
        return self.cluster.logs(self)

    def __getattr__(self, method):
        async def magic_rpc(*args, **kwargs):
            if len(args) > 0:
                raise TypeError('Positional arguments are not supported for RPC methods')
            return await self.call(method, kwargs)
        return magic_rpc

    def __str__(self):
        return f'RemoteTask({self.id}, {self.status}, {self.inputs})'

    def __repr__(self):
        return self.__str__()



from docker.types import Mount


def create_volumes(volumes: dict) -> list:
    mounts = [
        # this is the secret sauce that allows us to create new
        # tasks as sibling containers
        create_bind_mount('/var/run/docker.sock', {
            'src': '/var/run/docker.sock',
            'mode': 'ro',
        })
    ]

    for target, volume in volumes.items():
        if 'bind' in volume:
            mounts.append(create_bind_mount(target, volume['bind']))
        elif 'tmpfs' in volume:
            mounts.append(create_tmpfs_mount(target, volume['tmpfs']))

    return mounts


def create_bind_mount(target: str, bind) -> Mount:
    mode = 'rw'
    src = None
    if isinstance(bind, str):
        src = bind
    elif isinstance(bind, dict):
        src = bind.get('src')
        mode = bind.get('mode', 'rw')
    else:
        raise TypeError(f'Invalid bind volume definition {target}')

    return Mount(
        type='bind',
        target=target,
        source=src,
        read_only=mode != 'rw',
    )


def create_tmpfs_mount(target: str, tmpfs) -> Mount:
    size = 0
    mode = 0
    if isinstance(tmpfs, dict):
        size = tmpfs.get('size')
        mode = tmpfs.get('mode', 1777)
    else:
        raise TypeError(f'Invalid tmpfs volume definition {target}')

    return Mount(
        type='tmpfs',
        target=target,
        tmpfs_size=size,
        tmpfs_mode=mode,
    )



from kubernetes import client
from cowait.engine.const import LABEL_TASK_ID


def create_ports(ports: dict) -> list:
    return [
        client.V1ContainerPort(**convert_port(port, host_port))
        for port, host_port in ports.items()
    ]


def convert_port(port, host_port: str = None):
    protocol = 'TCP'
    if isinstance(port, str) and '/' in port:
        slash = port.find('/')
        protocol = port[slash+1:]
        port = port[:slash]

    if host_port is None:
        host_port = port

    return {
        'protocol': protocol.upper(),
        'container_port': int(port),
        'host_port': int(host_port),
    }




import click
from cowait.version import version
from . import task, image, context
from .cluster import cluster


def new_cli_app():
    @click.group()
    @click.version_option(version)
    @click.pass_context
    def cli(ctx):
        pass

    # task commands
    cli.add_command(task.run)
    cli.add_command(task.kill)
    cli.add_command(task.ps)
    cli.add_command(task.rm)
    cli.add_command(task.agent)
    cli.add_command(task.notebook)
    cli.add_command(task.test)

    # image commands
    cli.add_command(image.build)
    cli.add_command(image.push)

    # context commands
    cli.add_command(context.new)

    # cluster commands
    cli.add_command(cluster)

    return cli



from __future__ import annotations
from datetime import datetime, timezone
from ..utils import uuid


def generate_task_id(name: str) -> str:
    if '.' in name:
        dot = name.rfind('.')
        name = name[dot+1:]

    name = name.replace('.', '-')
    name = name.replace('_', '-')

    return '%s-%s' % (name.lower(), uuid())


class TaskDefinition(object):
    """
    Defines a Task :)

    Attributes:
        name (str): Task import name.
        image (str): Task image.
        id (str): Task id. If None, an id will be autogenerated.
        upstream (str): Upstream connection string. Defaults to None.
        inputs (dict): Input values
        meta (dict): Freeform metadata
        env (dict): Environment variables
        ports (dict): Port forwards
        routes (dict): HTTP Ingresses
        volumes (dict): List of volumes
        cpu (str): CPU request
        cpu_limit (str): CPU limit
        memory (str): Memory request
        memory_limit (str): Memory limit
        owner (str): Owner name
        created_at (DateTime): Creation date
    """

    def __init__(
        self,
        name:         str,
        image:        str,
        id:           str = None,
        upstream:     str = None,
        parent:       str = None,
        inputs:       dict = {},
        meta:         dict = {},
        env:          dict = {},
        ports:        dict = {},
        routes:       dict = {},
        volumes:      dict = {},
        cpu:          str = None,
        cpu_limit:    str = None,
        memory:       str = None,
        memory_limit: str = None,
        affinity:     str = None,
        owner:        str = '',
        created_at:   datetime = None,
    ):
        """
        Arguments:
            name (str): Task import name.
            image (str): Task image.
            id (str): Task id. If None, an id will be autogenerated.
            upstream (str): Upstream connection string. Defaults to None.
            inputs (dict): Input values
            meta (dict): Freeform metadata
            env (dict): Environment variables
            ports (dict): Port forwards
            routes (dict): HTTP Ingresses
            volumes (dict): List of volumes
            cpu (str): CPU request
            cpu_limit (str): CPU limit
            memory (str): Memory request
            memory_limit (str): Memory limit
            affinity (str): Affinity Mode (None/stack/spread)
            owner (str): Owner name
            created_at (DateTime): Creation date
        """
        self.id = generate_task_id(name) if id is None else id
        self.name = name
        self.image = image
        self.parent = parent
        self.upstream = upstream
        self.inputs = inputs
        self.meta = meta
        self.env = {str(k): str(v) for k, v in env.items()}
        self.ports = ports
        self.routes = routes
        self.cpu = cpu
        self.cpu_limit = cpu_limit
        self.memory = memory
        self.memory_limit = memory_limit
        self.owner = owner
        self.volumes = volumes
        self.affinity = affinity

        if created_at is None:
            self.created_at = datetime.now(timezone.utc)
        elif isinstance(created_at, datetime):
            self.created_at = created_at
        elif isinstance(created_at, str):
            self.created_at = datetime.fromisoformat(created_at)
        else:
            raise TypeError(f'Expected created_at to be None or datetime, got {created_at}')

    def serialize(self) -> dict:
        """ Serialize task definition to a dict """
        return {
            'id': self.id,
            'name': self.name,
            'image': self.image,
            'upstream': self.upstream,
            'parent': self.parent,
            'inputs': self.inputs,
            'meta': self.meta,
            'env': self.env,
            'ports': self.ports,
            'routes': self.routes,
            'cpu': self.cpu,
            'cpu_limit': self.cpu_limit,
            'memory': self.memory,
            'memory_limit': self.memory_limit,
            'affinity': self.affinity,
            'owner': self.owner,
            'created_at': self.created_at.isoformat(),
            'volumes': self.volumes,
        }

    @staticmethod
    def deserialize(taskdef: dict) -> TaskDefinition:
        """ Deserialize task definition from a dict """
        return TaskDefinition(
            id = taskdef.get('id'),
            name = taskdef.get('name'),
            image = taskdef.get('image'),
            upstream = taskdef.get('upstream', None),
            parent = taskdef.get('parent', None),
            inputs = taskdef.get('inputs', {}),
            meta = taskdef.get('meta', {}),
            env = taskdef.get('env', {}),
            ports = taskdef.get('ports', {}),
            routes = taskdef.get('routes', {}),
            cpu = taskdef.get('cpu', None),
            cpu_limit = taskdef.get('cpu_limit', None),
            memory = taskdef.get('memory', None),
            memory_limit = taskdef.get('memory_limit', None),
            affinity = taskdef.get('affinity', None),
            owner = taskdef.get('owner', None),
            created_at = datetime.fromisoformat(taskdef.get('created_at', datetime.now().isoformat())),
            volumes = taskdef.get('volumes', {}),
        )




import sys
import inspect
from typing import Any
from cowait.types import serialize
from .definition import TaskDefinition
from .components import TaskManager, RpcComponent, rpc
from .parent_task import ParentTask


class Task(object):
    __current__ = None

    def __init__(self, **inputs):
        """
        Creates a new instance of the task. Pass inputs as keyword arguments.
        """

        # We are using **inputs keyword arguments so that in-IDE tooltips will be more helpful
        # (at least less confusing) when invoking subtasks using constructor syntax.
        # However, subtasks will actually never be instantiated. The constructor call is
        # diverted by the runtime in Task.__new__().
        # Tasks should only be constructed by the executor, and it will these 3 arguments:
        if 'taskdef' not in inputs or 'node' not in inputs or \
           'cluster' not in inputs or len(inputs) != 3:
            raise RuntimeError('Invalid task class instantiation')

        self.node = inputs['node']
        self.cluster = inputs['cluster']
        self.taskdef = inputs['taskdef']

        self.parent = ParentTask(self.node)
        self.subtasks = TaskManager(self)
        self.rpc = RpcComponent(self)

        # Set this task as the current active task
        Task.set_current(self)

    def __new__(cls, *args, **inputs):
        current = Task.get_current()
        if current is None:
            # There is no active task. Continue normal instantiation.
            return object.__new__(cls)
        else:
            # There is already an active task in this process, so we should spawn a subtask.
            # Divert constructor behaviour to instead spawn a remote task and return it.

            if len(args) > 0:
                raise TypeError('Tasks do not accept positional arguments')

            return current.spawn(cls, inputs=inputs)

    @property
    def id(self) -> str:
        return self.taskdef.id

    @property
    def image(self) -> str:
        return self.taskdef.image

    @property
    def meta(self) -> dict:
        return self.taskdef.meta

    def __str__(self) -> str:
        return f'Task({self.id}, {self.name})'

    def init(self):
        pass

    async def before(self, inputs: dict) -> dict:
        return inputs

    async def run(self, **inputs: dict) -> Any:
        pass

    async def after(self, inputs: dict) -> Any:
        pass

    @rpc
    async def stop(self) -> None:
        """
        Abort task execution.
        """
        print('\n~~ STOPPED ~~')

        # send a stop status
        await self.node.parent.send_stop()

        # stop subtasks
        for task in self.subtasks.values():
            await task.stop()

        # schedule exit on the next event loop.
        # allows the RPC call to return before exit.
        async def _quit():
            sys.exit(1)
        self.node.io.create_task(_quit())

    def spawn(
        self,
        name: str,
        id: str = None,
        image: str = None,
        ports: dict = {},
        routes: dict = {},
        inputs: dict = {},
        meta: dict = {},
        env: dict = {},
        volumes: dict = {},
        cpu: str = None,
        cpu_limit: str = None,
        memory: str = None,
        memory_limit: str = None,
        affinity: str = None,
        owner: str = '',
        **kwargs: dict,
    ) -> 'Task':
        """
        Spawn a subtask.

        Arguments:
            name (str): Task name
            image (str): Task image. Defaults to the current task image.
            kwargs (dict): Input arguments
        """

        # merge inputs with remaining kwargs
        inputs = {
            **inputs,
            **kwargs,
        }

        if isinstance(name, str):
            pass
        elif issubclass(name, Task):
            name = name.__module__
        else:
            raise TypeError('Unsupported task type: ' + type(name))

        # throw error if any input is a coroutine
        for key, value in inputs.items():
            if inspect.iscoroutine(value):
                raise TypeError(f'Input {key} must be awaited first')

        taskdef = TaskDefinition(
            id=id,
            name=name,
            parent=self.id,
            image=image if image else self.image,
            upstream=self.node.get_url(),
            meta=meta,
            ports=ports,
            routes=routes,
            cpu=cpu if cpu else self.taskdef.cpu,
            cpu_limit=cpu_limit if cpu_limit else self.taskdef.cpu_limit,
            memory=memory if memory else self.taskdef.memory,
            memory_limit=memory_limit if memory_limit else self.taskdef.memory_limit,
            affinity=affinity if affinity else self.taskdef.affinity,
            owner=owner,
            inputs=serialize(inputs),
            volumes={
                **self.taskdef.volumes,
                **volumes,
            },
            env={
                **self.taskdef.env,
                **env,
            },
        )

        # authorize id
        self.node.server.auth.add_token(taskdef.id)

        # spawn task
        task = self.cluster.spawn(taskdef)

        # register with subtask manager
        self.subtasks.watch(task)

        return task

    @staticmethod
    def get_current() -> 'Task':
        return Task.__current__

    @staticmethod
    def set_current(task: 'Task'):
        Task.__current__ = task



from cowait.tasks import Task, rpc
from .rpc_child import RpcChild


class RpcParent(Task):
    def init(self):
        self.called = False

    async def run(self):
        child = RpcChild()
        print('waiting for child init...')
        await child.wait_for_init()

        print('calling child rpc')
        rpc_result = await child.echo(value=1)
        assert rpc_result['value'] == 1

        print('all done')
        task_result = await child
        assert task_result['called']

        # expect to receive rpc call from child
        assert self.called

    @rpc
    async def set_called(self) -> int:
        self.called = True
        return 1337





class ParentTask(object):
    def __init__(self, node):
        self.parent = node.parent

    @property
    def id(self) -> str:
        return self.parent.id

    async def call(self, method, args={}):
        if not self.parent.connected:
            raise RuntimeError('Not connected to parent')

        if self.parent.rpc is None:
            raise RuntimeError('Parent RPC unavailable')

        return await self.parent.rpc.call(method, args)

    def __getattr__(self, method):
        async def magic_rpc(*args, **kwargs):
            if len(args) > 0:
                raise TypeError('Positional arguments are not supported for RPC methods')
            return await self.call(method, kwargs)
        return magic_rpc



import os
import sys
from signal import signal, SIGINT
from .const import CONTEXT_FILE_NAME

HEADER_WIDTH = 80

try:
    rows, columns = os.popen('stty size', 'r').read().split()
    HEADER_WIDTH = int(columns)
except Exception:
    pass


class ExitTrap():
    def __init__(self, callback: callable, single=True):
        self.callback = callback
        self.prev_handler = None
        self.single = single

    def __enter__(self):
        self.attach(self.callback)

    def __exit__(self, *exc):
        self.reset()

    def attach(self, callback):
        def handler(a, b):
            if self.single:
                self.reset()
            callback()

        self.reset()
        old = signal(SIGINT, handler)
        self.prev_handler = old

    def reset(self):
        if self.prev_handler is not None:
            signal(SIGINT, self.prev_handler)
            self.prev_handler = None


def find_file_in_parents(start_path, file_name):
    """ Finds a file in a directory or any of its parent directories. """

    if not os.path.isdir(start_path):
        raise FileNotFoundError(f'Start directory not found at {start_path}')

    check_path = start_path
    while True:
        files = os.listdir(check_path)

        # if the file is found within the current directory, return its path
        if file_name in files:
            return os.path.join(check_path, file_name)

        # we have reached the root of the task context
        # if the file is not found by now - it doesn't exist
        if CONTEXT_FILE_NAME in files:
            return None

        # reached the file system root, abort mission
        # this will happen when the tool is run outside a task context
        if check_path == '/':
            return None

        # goto parent directory and keep looking
        check_path = os.path.dirname(check_path)


def printheader(title: str = None) -> None:
    if title is None:
        print(f'--'.ljust(HEADER_WIDTH, '-'))
    else:
        print(f'-- {title} '.upper().ljust(HEADER_WIDTH, '-'))


class ProgressBar(object):
    def __init__(self, id, label, minimum, maximum):
        self.id = id
        self.label = label
        self.value = 0
        self.minimum = minimum
        self.maximum = maximum

    def set(self, value):
        self.value = max(min(value, self.maximum), self.minimum)

    def print(self, label_width=32, bar_width=32, left='[', bar='=', right=']'):
        sys.stdout.write(self.label)
        sys.stdout.write(' ' * (label_width - len(self.label)))
        p = round((self.value / self.maximum) * bar_width) if self.maximum > 0 else 0
        sys.stdout.write(left + bar * p + ' ' * (bar_width - p) + right + ' ')
        sys.stdout.write(str(self.value) + '/' + str(self.maximum))
        sys.stdout.write('\n')


class ProgressBars(object):
    def __init__(self):
        self.bars = {}
        self.bar_width = 32
        self.label_width = 32

    def add(self, id, label, minimum, maximum):
        self.bars[id] = ProgressBar(id, label, minimum, maximum)
        if len(label) > self.label_width:
            self.label_width = len(label)
        sys.stdout.write('\n')

    def set(self, id, value, maximum=None):
        self.bars[id].set(value)
        if maximum is not None:
            self.bars[id].maximum = maximum

    def has(self, id):
        return id in self.bars

    def refresh(self):
        sys.stdout.write(u"\u001b[1000D") # Move left
        sys.stdout.write(u"\u001b[" + str(len(self.bars)) + "A") # Move up
        for bar in self.bars.values():
            bar.print(
                label_width=self.label_width, 
                bar_width=self.bar_width,
            )
        sys.stdout.flush()




# flake8: noqa: F401

from .run import run
from .build import build
from .push import push
from .agent import agent
from .new_context import new_context
from .test import test
from .notebook import notebook
from .cluster import *

from .misc import *



import time
import asyncio
from threading import Thread


class IOThread(Thread):
    def __init__(self):
        super().__init__()
        self.loop = None

    def run(self):
        # create a new event loop and set it as default loop for this thread
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()

    def start(self):
        # start thread and block until the event loop is ready
        super().start()
        while self.loop is None:
            time.sleep(0.01)

    def create_task(self, coro):
        future = asyncio.run_coroutine_threadsafe(coro, loop=self.loop)
        return asyncio.wrap_future(future)



from .run import run as run_cmd


def notebook(config, build: bool, image: str = None, cluster_name: str = None) -> None:
    if image is None:
        image = 'cowait/notebook'

    task = f'{image}/cowait.notebook'
    return run_cmd(
        config=config,
        task=task,
        build=build,
        routes={
            '/': '8888',
        },
        cluster_name=cluster_name,
    )



import pytest
import asyncio
from alt_pytest_asyncio.plugin import AltPytestAsyncioPlugin
from cowait import Task


class PytestTask(Task):
    async def run(self):
        loop = asyncio.get_event_loop()

        plugins = [
            AltPytestAsyncioPlugin(loop=loop),
        ]

        code = pytest.main(["-c", "/var/cowait/pytest.ini"], plugins=plugins)
        if code != pytest.ExitCode.OK and \
           code != pytest.ExitCode.NO_TESTS_COLLECTED:
            raise RuntimeError('Tests failed')



import asyncio
from cowait.tasks import Task, sleep, rpc
from cowait.tasks.messages import TASK_LOG
from concurrent.futures import Future
from dask.distributed import Client as DaskClient
from .worker import DaskWorker
from .scheduler import DaskScheduler
from .types import DaskClientType  # noqa: F401

MSG_LEADER = 'Scheduler at:'
MSG_REGISTER = 'Registered to:'


class DaskCluster(Task):
    def init(self):
        self.scheduler = None
        self.workers = []
        self.running = False

        # subscribe to logs
        self.node.server.on(TASK_LOG, self.on_log)

    async def on_log(self, id, file, data, **msg) -> None:
        if self.scheduler and id == self.scheduler.id:
            if MSG_LEADER in data:
                print('~~ dask scheduler ready')
                self.scheduler.ready.set_result(self.scheduler)

        for worker in self.workers:
            if id == worker.id and MSG_REGISTER in data:
                print(f'~~ dask worker {worker.id} ready')
                worker.ready.set_result(worker)

    async def before(self, inputs: dict) -> dict:
        await self.create_cluster(**inputs)
        return inputs

    async def run(self, workers: int = 5):
        print('~~ dask cluster running')
        self.running = True
        while self.running:
            await sleep(1)
        return 'ok'

    async def after(self, inputs: dict):
        await self.dask.close()

        print('~~ stopping dask cluster')
        self.scheduler.destroy()

        for worker in self.workers:
            worker.destroy()

        await super().after(inputs)

    async def create_cluster(self, workers=1) -> None:
        print('~~ creating dask cluster...')
        print(f'~~   workers = {workers}')

        # create dask scheduler
        self.scheduler = self.spawn(
            DaskScheduler,
            routes={
                # '/': 8787,
            },
        )
        self.scheduler.ready = Future()
        self.scheduler_uri = f'tcp://{self.scheduler.ip}:8786'

        await sleep(1)

        # create workers
        self.workers = []
        await self.add_workers(workers)

        print('~~ waiting for cluster nodes')
        await self.wait_for_nodes()

        print('~~ dask cluster ready')
        self.dask = DaskClient(address=self.scheduler_uri)

    async def add_workers(self, count):
        for i in range(0, count):
            w = DaskWorker(scheduler=self.scheduler_uri)
            w.ready = Future()
            self.workers.append(w)

    async def remove_workers(self, count):
        count = abs(count)
        if count > len(self.workers):
            count = len(self.workers)

        for worker in self.workers[-count:]:
            worker.destroy()
        self.workers = self.workers[:-count]

    async def wait_for_nodes(self):
        await asyncio.gather(
            asyncio.wrap_future(self.scheduler.ready),
            *map(lambda w: asyncio.wrap_future(w.ready), self.workers),
        )

    @rpc
    async def get_workers(self):
        ready = filter(lambda w: w.ready.done(), self.workers)
        ids = list(map(lambda w: w.id, ready))
        return {
            'workers': ids,
            'ready': len(ids),
            'total': len(self.workers),
        }

    @rpc
    async def scale(self, workers: int):
        diff = workers - len(self.workers)
        if diff > 0:
            # scale up
            print(f'~~ scale({workers}): adding {diff} workers')
            await self.add_workers(diff)

        elif diff < 0:
            # scale down
            print(f'~~ scale({workers}): removing {abs(diff)} workers')
            await self.remove_workers(diff)

    @rpc
    async def get_scheduler_uri(self) -> str:
        return self.scheduler_uri

    @rpc
    async def get_client(self) -> DaskClient:
        """ Returns a Dask client for this cluster """
        return self.dask

    @rpc
    async def teardown(self) -> None:
        self.running = False



import sys
import json
from ..task_image import TaskImage
from ..const import CONTEXT_FILE_NAME
from ..config import Config
from ..context import Context
from ..utils import ProgressBars


def push(config: Config, **kwargs) -> TaskImage:
    context = Context.open(config)
    image = TaskImage.open(context)

    if image is None:
        print('Error: Failed to resolve task image')
        return

    if '/' not in image.name:
        print(f'Error: You must specify a full image name in '
              f'{CONTEXT_FILE_NAME} before you can push')
        return

    print('Pushing image')

    progress = ProgressBars()
    logs = image.push()
    for update in logs:
        if 'id' in update and 'progressDetail' in update:
            id = update['id']
            detail = update['progressDetail']
            if not progress.has(id):
                progress.add(id, id, 0, 0)
            if 'total' in detail:
                progress.set(id, detail['current'], detail['total'])
            progress.refresh()

        if 'errorDetail' in update:
            print('Error:', update['error'])
            return None

    return image



from .errors import PodTerminatedError, PodUnschedulableError, ImagePullError


def pod_is_ready(pod):
    if pod.status.conditions:
        for cond in pod.status.conditions:
            if cond.type == 'PodScheduled' and cond.status == 'False':
                raise PodUnschedulableError(cond.message)

    statuses = pod.status.container_statuses
    if statuses is None:
        return False

    for status in statuses:
        state = status.state

        if state.running is not None:
            return True

        if state.terminated is not None:
            raise PodTerminatedError()

        if state.waiting is not None:
            if state.waiting.reason == 'ContainerCreating':
                return False

            if state.waiting.reason == 'ErrImagePull':
                raise ImagePullError(state.waiting.message)

            if state.waiting.reason == 'ImagePullBackoff':
                raise ImagePullError(state.waiting.message)

    return False




from .type import Type


class CustomType(Type):
    def __init__(self, cls, shape: dict):
        self.cls = cls
        self.shape = shape

    def validate(self, value):
        if not isinstance(value, dict):
            raise TypeError('Expected a dict')
        for key, type in self.shape.items():
            type.validate(value[key])

    def serialize(self, value):
        return {
            key: type.serialize(getattr(value, key))
            for key, type in self.shape.items()
        }

    def deserialize(self, value):
        return self.cls(**{
            key: type.deserialize(value[key])
            for key, type in self.shape.items()
        })



import aiohttp
import asyncio
import requests
import traceback
from threading import Thread
from queue import Queue, Empty as QueueEmpty
from cowait.tasks import TaskDefinition, RemoteTask
from cowait.tasks.messages import TASK_FAIL, TASK_RETURN, TASK_LOG
from cowait.tasks.components.rpc import RpcError
from .cluster import ClusterProvider
from .errors import TaskCreationError


class ApiProvider(ClusterProvider):
    def __init__(self, args={}):
        super().__init__('api', args)
        self.url = args.get('url')
        self.token = args.get('token', None)

    def rpc(self, method: str, **kwargs) -> dict:
        # todo: authentication
        url = f'{self.url}/rpc/{method}'
        resp = requests.post(url, json={'args': kwargs}, headers={
            'Cowait-Key': self.token,
        })
        if resp.status_code == 401:
            raise RuntimeError('Authentication error. Invalid token?')

        msg = resp.json()
        if resp.status_code == 200 and 'result' in msg:
            return msg['result']
        if resp.status_code == 400 and 'error' in msg:
            raise RpcError(msg['error'])
        raise RpcError(f'Request status {resp.status_code}')

    def spawn(self, taskdef: TaskDefinition) -> RemoteTask:
        try:
            task = self.rpc('spawn', **taskdef.serialize())
            return RemoteTask(TaskDefinition.deserialize(task), self)
        except RpcError as e:
            raise TaskCreationError(str(e))

    def destroy(self, task_id):
        self.rpc('destroy', task_id=task_id)

    def destroy_all(self):
        self.rpc('destroy_all')

    def list_all(self):
        return self.rpc('list_tasks')

    def logs(self, task):
        ws_url = self.args.get('ws_url')
        if ws_url is None:
            print('No websocket URL set - logs not available.')
            return

        watcher = ApiLogsWatcher(task.id, ws_url)
        for log in watcher.watch():
            yield log

    def find_agent(self):
        return self.rpc('get_agent_url')


class ApiLogsWatcher(Thread):
    END = object()

    def __init__(self, task_id, ws_url):
        super().__init__()
        self.task_id = task_id
        self.ws_url = ws_url
        self.queue = Queue()

    def run(self):
        loop = asyncio.new_event_loop()
        loop.run_until_complete(self.__watch())

    def watch(self, timeout=60):
        self.start()
        while True:
            try:
                log = self.queue.get(True, timeout)
                if log is self.END:
                    break
                yield log
            except QueueEmpty:
                break

    async def __watch(self):
        try:
            session = aiohttp.ClientSession()
            ws = await session.ws_connect(self.ws_url)
            await ws.send_json({
                'type': 'subscribe'
            })
            while True:
                msg = await ws.receive()
                if msg.type == aiohttp.WSMsgType.CLOSE:
                    self.queue.put('!! Lost API connection')
                    break
                if msg.type == aiohttp.WSMsgType.ERROR:
                    self.queue.put('!! Lost API connection')
                    break

                event = msg.json()
                if 'id' not in event or event['id'] != self.task_id:
                    continue
                if event['type'] == TASK_RETURN:
                    break
                if event['type'] == TASK_FAIL:
                    self.queue.put(f'!! {self.task_id} failed with error:')
                    self.queue.put(event['error'].strip())
                    break
                if event['type'] == TASK_LOG:
                    self.queue.put(event['data'].strip())

        except Exception:
            traceback.print_exc()
        finally:
            await ws.close()
            await session.close()

        self.queue.put(self.END)



from ..config import Config
from ..context import Context
from cowait.engine import ProviderError


def destroy(config: Config, *, cluster_name: str = None) -> None:
    try:
        context = Context.open(config)
        cluster = context.get_cluster(cluster_name)
        cluster.destroy_all()

    except ProviderError as e:
        print('Provider error:', str(e))


def list_tasks(config: Config, *, cluster_name: str = None) -> None:
    try:
        context = Context.open(config)
        cluster = context.get_cluster(cluster_name)
        tasks = cluster.list_all()
        for task in tasks:
            print(task)

    except ProviderError as e:
        print('Provider error:', str(e))


def kill(config: Config, partial_task_id: str):
    """ Kills all tasks that somewhat matches partial_task_id """
    try:
        cluster = config.get_cluster()
        tasks = cluster.list_all()
        for task in tasks:
            if partial_task_id in str(task):
                cluster.destroy(str(task))
                print('killed ', str(task))

    except ProviderError as e:
        print('Provider error:', str(e))



import yaml


class SettingsDict(object):
    def __init__(self, *, path: str = None, data: dict = None, parent: 'SettingsDict' = None):
        if path is not None:
            if data is not None:
                raise ValueError('Pass either path or a data dict')
            self.read(path)
        else:
            self.data = {} if data is None else data

        self.parent = parent

    def __getitem__(self, key: str) -> any:
        return self.get(key, required=True)

    def __setitem__(self, key: str, value: any):
        self.set(key, value)

    def __delattr__(self, key: str):
        self.delete(key)

    def set(self, key: str, value: any) -> any:
        path = split_key(key)
        container = self.data
        for part in path[:-1]:
            if part not in container:
                container[part] = {}
            if not isinstance(container, dict):
                raise TypeError(f'{part} is not a dict')
            container = container[part]

        container[path[-1]] = value
        return value

    def override(self, key: str, value: any) -> any:
        if value is None:
            return self.get(key, None, False)
        return self.set(key, value)

    def has(self, key: str) -> bool:
        path = split_key(key)
        container = self.data
        for part in path[:-1]:
            if part not in container:
                if self.parent:
                    return self.parent.has(key)
                return False
            if not isinstance(container, dict):
                return False
            container = container[part]
        return path[-1] in container

    def get(self, key: str, default: any = None, required: bool = True) -> any:
        path = split_key(key)
        value = self.data
        for part in path:
            if not isinstance(value, dict) or part not in value:
                value = None
                break
            value = value.get(part, default)
        if value is None:
            if self.parent:
                return self.parent.get(key, default, required)
            if default is None and required:
                raise KeyError(f'{key} not set')
            return default
        return value

    def delete(self, key: str) -> None:
        # TODO: propagate delete to parents
        path = split_key(key)
        container = self.data
        for part in path[:-1]:
            if part not in container:
                container[part] = {}
            if not isinstance(container, dict):
                raise TypeError(f'{part} is not a dict')
            container = container[part]
        del container[path[-1]]

    def coalesce(self, key: str, value: any, default: any) -> any:
        if value is not None:
            return value
        return self.get(key, default, required=False)

    def write(self, path):
        with open(path, 'w') as f:
            yaml.dump(
                self.pack_data(self.data),
                stream=f,
                sort_keys=False,
                default_flow_style=False,
            )

    def read(self, path):
        with open(path, 'r') as f:
            self.data = self.unpack_data(yaml.load(f, Loader=yaml.FullLoader))

    def pack_data(self, data):
        return data

    def unpack_data(self, data):
        return data


def split_key(key):
    path = None
    if isinstance(key, str):
        path = key.split('.')
    elif isinstance(key, list):
        path = key
    else:
        raise TypeError("Expected key to be str or list")

    if len(path) < 1:
        raise ValueError(f'Invalid key {key}')

    return path



import json
import re

identifier = re.compile(r"^[^\d\W]\w*\Z", re.UNICODE)


def parse_input(value: str) -> tuple:
    if '=' not in value:
        raise ValueError('Input is not on key=value format')

    split = value.index('=')
    key = value[:split].strip()
    value = value[split+1:].strip()

    if identifier.match(key) is None:
        raise ValueError('Input key must be a valid identifier')

    if len(value) == 0:
        raise ValueError('Input value can not be empty')

    try:
        value = json.loads(value)
    except json.JSONDecodeError:
        pass

    return key, value


def parse_input_list(inputs: list) -> dict:
    output = {}
    for input in inputs:
        key, value = parse_input(input)
        output[key] = value
    return output




from asyncio import CancelledError
from aiohttp import web, WSMsgType
from aiohttp_middlewares import cors_middleware
from datetime import datetime
from cowait.utils import EventEmitter
from .conn import Conn
from .const import WS_PATH, ON_CONNECT, ON_CLOSE, ON_ERROR
from .auth_middleware import AuthMiddleware
from .errors import SocketError


class Server(EventEmitter):
    def __init__(self, port, middlewares: list = []):
        super().__init__()
        self.conns = []
        self.port = port
        self.auth = AuthMiddleware()

        # create http app
        self.app = web.Application(
            middlewares=[
                *middlewares,
                self.auth.middleware,
                cors_middleware(allow_all=True)
            ],
        )

        # route shortcuts
        self.add_routes = self.app.router.add_routes
        self.add_route = self.app.router.add_route
        self.add_post = self.app.router.add_post
        self.add_get = self.app.router.add_get

        self.add_get(f'/{WS_PATH}', self.handle_client)

    async def handle_client(self, request):
        ws = web.WebSocketResponse(
            timeout=30.0,
            autoping=True,
            heartbeat=5.0,
        )
        await ws.prepare(request)

        conn = Conn(ws, request.remote)
        await self.emit(type=ON_CONNECT, conn=conn)
        self.conns.append(conn)

        try:
            while not ws.closed:
                msg = await ws.receive()
                if msg.type == WSMsgType.CLOSE:
                    break
                elif msg.type == WSMsgType.ERROR:
                    raise SocketError(ws.exception())
                elif msg.type == WSMsgType.BINARY:
                    raise SocketError('Unexpected binary message')

                event = msg.json()
                if conn.rpc.intercept_event(**event):
                    continue
                await self.emit(**event, conn=conn)

            await self.emit(type=ON_CLOSE, conn=conn)

        except CancelledError as e:
            raise e

        except SocketError as e:
            await self.emit(type=ON_ERROR, conn=conn, error=str(e))

        finally:
            # disconnected
            self.conns.remove(conn)

    async def send(self, msg: dict) -> None:
        msg['ts'] = datetime.now().isoformat()
        if 'type' not in msg:
            raise Exception('Messages must have a type field')

        for ws in self.conns:
            await ws.send_json(msg)

    async def serve(self) -> None:
        self._runner = web.AppRunner(self.app, handle_signals=False)
        await self._runner.setup()

        site = web.TCPSite(self._runner, host='0.0.0.0', port=self.port)
        await site.start()

    async def close(self) -> None:
        for conn in self.conns:
            await conn.close()

        await self._runner.cleanup()
        self._runner = None



"""
Task worker entry point.

Environment:
    COWAIT_CLUSTER (json): JSON-serialized ClusterProvider
    COWAIT_TASK (json): JSON-serialized TaskDefinition
"""
import os
import sys
import json
import asyncio
import traceback
import nest_asyncio
from datetime import datetime
from cowait.tasks.messages import TASK_FAIL
from cowait.worker import execute, \
    env_get_cluster_provider, \
    env_get_task_definition


async def main():
    # add working directory to pythonpath
    sys.path += [os.getcwd()]

    # unpack cluster provider
    cluster = env_get_cluster_provider()

    # unpack task definition
    taskdef = env_get_task_definition()

    # execute task
    try:
        await execute(cluster, taskdef)
        await asyncio.sleep(0.1)

    except Exception:
        print(json.dumps({
            'id': taskdef.id,
            'type': TASK_FAIL,
            'error': traceback.format_exc(),
            'ts': datetime.now().isoformat(),
        }))
        os._exit(1)

    os._exit(0)


if __name__ == "__main__":
    # apply a patch that allows nested asyncio loops
    nest_asyncio.apply()

    # run asyncio loop
    asyncio.run(main())

else:
    raise ImportError('Worker.__main__ is not meant to be imported')



from ..config import Config


ADDABLE_PROVIDERS = {
    'api': ['url', 'token'],
    'kubernetes': ['context'],
}


def cluster_get(config: Config, name: str) -> None:
    if name not in config.clusters:
        print('Unknown cluster', name)
        return 1

    args = config.get(['clusters', name])
    print(name)
    if name == config.default_cluster:
        print('    default')
    for key, value in args.items():
        print(f'    {key}: {value}')


def cluster_ls(config: Config) -> None:
    for name in config.clusters:
        cluster_get(config, name)
        print()


def cluster_add(config: Config, name: str, type: str, **options) -> None:
    if name in config.clusters:
        print(f'Error: Cluster {name} already exists')
        return 1

    if type not in ADDABLE_PROVIDERS:
        print(f'Error: Cant add cluster of type {type}')
        return 1

    required_fields = ADDABLE_PROVIDERS[type]
    for field in required_fields:
        if field not in options:
            print(f'Error: {field} option must be provided when creating {type} clusters.')
            return 1

    config.set(['clusters', name], {
        'type': type,
        **options,
    })
    config.write()

    # dump added cluster
    cluster_get(config, name)


def cluster_rm(config: Config, name: str) -> None:
    if name not in config.clusters:
        print(f'Error: Cluster {name} does not exist')
        return 1

    if name == config.default_cluster:
        print('Error: Cant remove the current default cluster')
        return 1

    if name == 'docker':
        print('Error: Cant remove the docker provider')
        return 1

    if name == 'kubernetes':
        print('Error: Cant remove the kubernetes provider')
        return 1

    config.delete(['clusters', name])
    config.write()


def cluster_default(config: Config) -> None:
    print(config.default_cluster)


def cluster_set_default(config: Config, name: str) -> None:
    if name not in config.clusters:
        print(f'Error: Cluster {name} does not exist')
        return 1

    config.set('default_cluster', name)
    config.write()



from setuptools import setup, find_packages

exec(open('cowait/version.py').read())

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='cowait',
    version=version,
    author='Backtick Technologies',
    description='Cowait is a framework for creating ' +
                'containerized workflows with asynchronous Python.',

    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://github.com/backtick-se/cowait',

    packages=find_packages(exclude=["test.*", "test"]),
    entry_points={
        'console_scripts': [
            'cowait = cowait.__main__:main',
        ],
    },

    classifiers=[],

    python_requires='>=3.7',
    install_requires=[
        'docker>=4',
        'kubernetes>=11,<13',
        'nest-asyncio>=1.4.1',
        'aiohttp>=3',
        'aiohttp-middlewares>=1',
        'pytest>=6',
        'alt-pytest-asyncio>=0.5.3',  # would be nice to move to the widely used pytest-asyncio
        'python-dotenv>=0.15',

        # cli
        'click>=7',
        'PyYAML>=5',
        'sty==1.0.0-beta.12',  # tty coloring, seems to be stuck in beta. remove/replace

        # utilities, not actually required
        'numpy >= 1',  # provides typing for numpy. unlikely to break
        'pytest-sugar >= 0.9',
        'pytest-cov >= 2',

        # extract to cowait-dask
        'dask >= 2',
        'distributed >= 2',
    ]
)



import sys
from cowait.tasks import TaskDefinition
from cowait.engine.errors import TaskCreationError, ProviderError
from cowait.utils.const import DEFAULT_BASE_IMAGE
from cowait.utils import uuid
from ..errors import CliError
from ..config import Config
from ..context import Context
from ..utils import ExitTrap
from .run import RunLogger


def agent(
    config: Config,
    detach: bool = False,
    upstream: str = None,
    cluster_name: str = None,
) -> None:
    logger = RunLogger(quiet=False, raw=False)
    try:
        context = Context.open(config)
        cluster = context.get_cluster(cluster_name)

        if cluster.type == 'api':
            raise CliError('Error: Cant deploy agent using an API cluster')

        token = uuid()
        if cluster.type == 'docker':
            token = ''

        cluster.destroy('agent')

        # create task definition
        taskdef = TaskDefinition(
            id='agent',
            name='cowait.tasks.agent',
            image=DEFAULT_BASE_IMAGE,
            upstream=upstream,
            routes={
                '/': 80,
            },
            meta={
                'http_token': token,
            },
        )

        # submit task to cluster
        task = cluster.spawn(taskdef)

        if detach:
            logger.header('detached')
            return

        def destroy(*args):
            logger.header('interrupt')
            cluster.destroy(task.id)
            sys.exit(0)

        with ExitTrap(destroy):
            # capture & print logs
            logs = cluster.logs(task)
            logger.header('task output')
            for log in logs:
                logger.handle(log)

        logger.header()

    except ProviderError as e:
        raise CliError(f'Provider error: {e}')

    except TaskCreationError as e:
        raise CliError(f'Task creation error: {e}')
